/***************************** LICENSE START ***********************************

 Copyright 2012 ECMWF and INPE. This software is distributed under the terms
 of the Apache License version 2.0. In applying this license, ECMWF does not
 waive the privileges and immunities granted to it by virtue of its status as
 an Intergovernmental Organization or submit itself to any jurisdiction.

 ***************************** LICENSE END *************************************/

#ifndef grbsh_DEFINED
#define grbsh_DEFINED

      INTEGER JMTRUNC,JMUAF,JMGRIB,JNS,JLEV
      PARAMETER (JMTRUNC=1023)			! TRUNCATION
      PARAMETER (JMUAF=(JMTRUNC+1)*(JMTRUNC+2))
      PARAMETER (JMGRIB=((JMUAF*3)/4)+1000)
      PARAMETER (JNS=150)			! MAXIMUM NUMBER OF STEPS
      PARAMETER (JLEV=24)			! NUMBER OF LEVELS

#endif
