/***************************** LICENSE START ***********************************

 Copyright 2013 ECMWF and INPE. This software is distributed under the terms
 of the Apache License version 2.0. In applying this license, ECMWF does not
 waive the privileges and immunities granted to it by virtue of its status as
 an Intergovernmental Organization or submit itself to any jurisdiction.

 ***************************** LICENSE END *************************************/


#ifndef MvQStationsHelp_H
#define MvQStationsHelp_H

#include "MvQPalette.h"
#include "MvQRequestPanelHelp.h"
#include "MvIconParameter.h"

#include <QWidget>

class MvQStationsWidget;

class RequestPanel;

class MvQStationsHelp : public MvQRequestPanelHelp
{
    Q_OBJECT

public:
    MvQStationsHelp(RequestPanel& owner, const MvIconParameter& param);
    ~MvQStationsHelp(){};

    void start(){};
    bool dialog() { return false; }
    QWidget* widget();

public slots:
    void slotSelected(QString);

protected:
    virtual void refresh(const vector<string>&);

private:
    MvQStationsWidget* selector_;
};

#endif
