/*
 * © Copyright 1996-2017 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef mars_client_hash_h
#define mars_client_hash_h

#if defined(__cplusplus) || defined(c_plusplus)
extern "C" {
#endif

/* hash.c */
void hash_stat(void);
char *strcache(const char *name);
void strfree(char *name);

#if defined(__cplusplus) || defined(c_plusplus)
}
#endif

#endif
