/*
 * (C) Copyright 2013 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#pragma once

#include "eckit/system/Plugin.h"

namespace atlas {

//----------------------------------------------------------------------------------------------------------------------

class Plugin : public eckit::system::Plugin {
public:
    using eckit::system::Plugin::Plugin;

protected:
    /// @brief Initialisation function called on loading the plugin dynamically
    void init() override;
};

//----------------------------------------------------------------------------------------------------------------------

}  // namespace atlas
