/*
 * (C) Copyright 2013 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

// TL31

#include "atlas/grid/detail/spacing/gaussian/N.h"

namespace atlas {
namespace grid {
namespace spacing {
namespace gaussian {

DEFINE_GAUSSIAN_LATITUDES( 16,
                           LIST( 85.7605871204438159, 80.2687790722500125, 74.7445403686357679, 69.2129761693708616,
                                 63.6786355610968613, 58.1429540492032828, 52.6065260343452650, 47.0696420596876806,
                                 41.5324612466560765, 35.9950784112715994, 30.4575539611520938, 24.9199286299486111,
                                 19.3822313464343878, 13.8444837343848572, 8.3067028565188039, 2.7689030077360099 ) )

}  // namespace gaussian
}  // namespace spacing
}  // namespace grid
}  // namespace atlas
