/***************************** LICENSE START ***********************************
 * Copyright 2012 ECMWF and INPE. This software is distributed under the terms
 * of the Apache License version 2.0. In applying this license, ECMWF does not
 * waive the privileges and immunities granted to it by virtue of its status as
 * an Intergovernmental Organization or submit itself to any jurisdiction.
 ***************************** LICENSE END *************************************/

 html
    {
        color: #000000;
        background-color: #FFFFFF;
    }
  
    /*table
    {
	border-color:  #0000FF;
	background-color: #FF0000;

     }*/

/*  table
   {
        background-color: #F6F6F6;
        border: 1px solid #000000;
        border-collapse: separate;
        font-size: 11px;
        margin-bottom: 25px; 
	border-collapse: collapse;
        border-spacing: 0;
    }*/
   td
    {
        padding: 3px 10px 3px 10px;
	background-color: #F9F9F9;
        color: #000000;
    }

   td.first
    {
        padding: 3px 10px 3px 10px;
	font-weight: bold;
	/*font-size: large;*/
	background-color: #F6F6F6;
        color: #000000;
    }
 
   td.image
    {
       background-color: #F9F9F9;
       color: #F9F9F9;
    }

   td.title
    {
	font-weight: bold;
	/*font-size: large;*/
	background-color: #E7F6FF;
        color: #000000;
    }
   td.highlight
    {
	/*font-size: large;*/
	background-color: #E7F6FF;
        color: #000000;
    }
  th
    {
       padding: 3px 10px 3px 10px;
       background-color: #F9F9F9;
       color: #000000;

    }


/*   tr
    {
        background-color: #F6F6F6;
        color: #000000;
    }
  
*/