/***************************** LICENSE START ***********************************

 Copyright 2012 ECMWF and INPE. This software is distributed under the terms
 of the Apache License version 2.0. In applying this license, ECMWF does not
 waive the privileges and immunities granted to it by virtue of its status as
 an Intergovernmental Organization or submit itself to any jurisdiction.

 ***************************** LICENSE END *************************************/

#ifndef grbll_DEFINED
#define grbll_DEFINED

C 1600/801 allows 0.225/0.225 lat/lon and N400 Gaussian

      INTEGER NMAXLON,NMAXLAT,JMUAF,JMGRIB
      PARAMETER (NMAXLON=1600,NMAXLAT=801)
      PARAMETER (JMUAF=NMAXLON*NMAXLAT)
      PARAMETER (JMGRIB=((JMUAF*3)/4)+1000)

#endif
