/***************************** LICENSE START ***********************************

 Copyright 2013 ECMWF and INPE. This software is distributed under the terms
 of the Apache License version 2.0. In applying this license, ECMWF does not
 waive the privileges and immunities granted to it by virtue of its status as
 an Intergovernmental Organization or submit itself to any jurisdiction.

 ***************************** LICENSE END *************************************/

#ifndef MvQFolderNavigation_H
#define MvQFolderNavigation_H

#include <QDebug>
#include <QStringList>

class MvQFolderNavigation
{
public:
    MvQFolderNavigation() :
        pos_(-1), maxSize_(300) {}

    void clear();
    bool isEmpty() const {return folders_.isEmpty();}
    bool hasNext();
    bool hasPrev();
    QString next();
    QString prev();
    QString current();
    void add(QString path);
    void removeAfterCurrent();
    void removeBeforeCurrent();
    void print();

protected:
    bool isPosCorrect();

    QStringList folders_;
    int pos_;
    int maxSize_;
};


#endif
