#Mass density (concentration)
'kg m**-3' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 0 ;
        }

#Mass mixing ratio
'kg kg**-1' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 2 ;
        }

#FLEXPART tracer mass concentration
'ng m**-3' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 192 ;
        }

#FLEXPART tracer mass ratio
'ppt by mass' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 193 ;
        }

#FLEXPART particle resident time
's' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 194 ;
        }

#FLEXPART response function for mass concentration
'm**3 s kg**-1' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 195 ;
        }

#FLEXPART response function for mass ratio
'kg s m**-3' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 196 ;
        }

#FLEXPART dry deposition concentration
'kg m**-2' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 197 ;
        }

#FLEXPART wet deposition concentration
'kg m**-2' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 198 ;
        }

#FLEXPART dry deposition mixing ratio
'kg kg**-1' = {
          discipline = 0 ;
          parameterCategory = 20 ;
          parameterNumber = 199 ;
         }

#FLEXPART wet deposition mixing ratio
'kg kg**-1' = {
          discipline = 0 ;
          parameterCategory = 20 ;
          parameterNumber = 200 ;
}

#FLEXPART eastward flux
'kg m**-2 s**-1' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 201 ;
        }

#FLEXPART westward flux
'kg m**-2 s**-1' = {
          discipline = 0 ;
          parameterCategory = 20 ;
          parameterNumber = 202 ;
         }

#FLEXPART southward flux
'kg m**-2 s**-1' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 203 ;
        }

#FLEXPART northward flux
'kg m**-2 s**-1' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 204 ;
        }

#FLEXPART upward flux
'kg m**-2 s**-1' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 205 ;
        }

#FLEXPART downward flux
'kg m**-2 s**-1' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 206 ;
        }


